"""
CannPatent | Cannlytics
Copyright (c) 2022 Cannlytics

Created: 5/21/2022
Updated: 6/5/2023
Authors: Keegan Skeate <https://github.com/keeganskeate>
License: MIT License <https://opensource.org/licenses/MIT>

Description:

    Find and curate data for cannabis patents. In particular, this
    script collects detailed data for plant patents. Subsequent
    intellectual property (IP) analytics provide actionable insights
    for cannabis cultivar inventors and consumers. For example,
    cultivators can use the methodology to predict if a particular
    cultivar would make a good patent candidate given its lab results.
    Consumers can find the nearest patented strain to a set of lab results
    printed on a cultivar's label.

Data Source:

    - United States Patent and Trademark Office
    URL: <www.uspto.gov>

Requirements

    - ImageMagick
    Download: <https://imagemagick.org/script/download.php#windows>
    Cloud: <https://stackoverflow.com/questions/43036268/do-i-have-access-to-graphicsmagicks-or-imagemagicks-in-a-google-cloud-function>

"""
# Standard imports:
from datetime import datetime
import math
import os
import re
from time import sleep
from typing import Any, List, Optional, Union

# External imports:
from bs4 import BeautifulSoup
import pandas as pd
from pydantic import BaseModel
import requests

# Internal imports:
from cannlytics.firebase.firebase import (
    initialize_firebase,
    update_document,
    update_documents,
)
from cannlytics.utils.utils import (
    camel_to_snake,
    clean_dictionary,
)


class StrainLabResults(BaseModel):
    """A data class representing a cannabis strain."""
    key: str
    avg: float
    min: float
    max: float
    median: float
    std: float
    sample_ids: List[str]


class Strain(BaseModel):
    """A data class representing a cannabis strain."""
    cultivar: str
    strain_name: str
    first_produced_at: str
    first_produced_by: str
    first_produced_state: str
    number_of_tests: int
    results: List[StrainLabResults]


class CannPatent():
    """Cultivar identifier."""
    pass


def search_patents(
        query: str,
        limit: Optional[int] = 50,
        details: Optional[bool] = False,
        pause: Optional[float] = None,
        term: Optional[str] = '',
    ) -> Any:
    """Search for patents.
    Args:
        query (str): The search term.
        limit (int): The number of patents to retrieve, 50 by default.
            The algorithm requests patents in batches of 50.
        details (bool): Whether or not to return extensive details for
            each patent. The default is `False`.
    Returns:
        (DataFrame): The patents' data.
    """

    # Define the query URL.
    query = query.replace(' ', '+')
    base = 'http://appft.uspto.gov/netacgi/nph-Parser'
    # base = 'https://developer.uspto.gov/ibd-api'
    headers = {'User-Agent': 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0'}
    url = f'{base}?Sect1=PTO2&Sect2=HITOFF&u=%2Fnetahtml%2FPTO%2Fsearch-adv.htm&r=0&p=1&f=S&l=50&Query='
    url += f'{term}"{query}"&d=PTXT'
    print(url)

    # Future work: Allow user to specify date ranges for the search query.
    # ISD/11/1/1997->5/12/1998

    # Iterate over found items, 50 per page.
    patents = pd.DataFrame()
    pages = math.ceil(limit / 50)
    page_range = range(0, int(pages))
    for increment_page in page_range[:1]:

        # Get cannabinoid patents.
        if increment_page:
            url += f'&OS={query}&RS={query}'
            url += f'&TD=6080&Srch1={query}&NextList{increment_page + 1}=Next+50+Hits'
        # FIXME:
        try:
            response = requests.get(url, headers=headers)
        except:
            print('Error on page', increment_page)
            sleep(62)
            response.connection.close()
            response = requests.get(url, headers=headers)
            # try:
            #     response = requests.get(url, headers=headers)
            # except:
            #     continue
        if pause:
            sleep(pause)

        # Create lists of patents, titles, and URLs.
        soup = BeautifulSoup(response.content, features='html.parser')
        ids = []
        patent_numbers = []
        titles = []
        links = []
        for link in soup.find_all('a', href=True):
            if link.text and (link.text.startswith('PP') or (len(link.text) <= 10 and ',' in link.text)):
                ids.append(link.text)
                patent_number = link.text.replace(',', '')
                patent_numbers.append(patent_number)
                title_link = link.findNext('a')
                titles.append(title_link.text.strip())
                links.append('http://patft.uspto.gov' + title_link['href'])

        # Format the patents as a DataFrame.
        if patent_numbers == 0:
            break
        patents = pd.concat([patents, pd.DataFrame({
            'patent_number': patent_numbers,
            'patent_number_formatted': ids,
            'patent_title': titles,
            'patent_url': links,
        })])

        # Optionally get details for each patent.
        # Note: This can probably be done more efficiently.
        if details:
            patent_details = []
            for _, patent in patents[:limit].iterrows():
                patent_detail = get_patent_details(patent)
                patent_details.append(patent_detail)
            patents = pd.concat(patent_details, axis=1)
            if isinstance(patents, pd.Series):
                patents = patents.to_frame()
            patents = patents.transpose()

    # Return the patents found.
    # FIXME: Only return the limit?
    # patents = patents[:limit]
    return patents


def get_patent_details(
        data: Optional[Any] = None,
        patent_number: Optional[str] = None,
        patent_url: Optional[str] = None,
        user_agent: Optional[str] = 'Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:52.0) Gecko/20100101 Firefox/52.0',
        fields: Optional[Union[str, list]] = None,
        search_field: Optional[str] = 'patentNumber',
        search_fields: Optional[str] = 'patentNumber',
        query: Optional[str] = 'patentNumber',
    ) -> Any:
    """Get details for a given patent, given it's patent number and URL.
    Args:
        data (Series): Existing patent data with `patent_number` and
            `patent_url` fields (optional). If not specified, then
            pass `patent_number` and `patent_url` arguments.
        patent_number (str): A specific patent number (optional).
        patent_url (str): A specific patent URL (optional).
        user_agent (str): Your browser agent (optional). A typical Chrome
            agent by default.
        fields (list): A list of fields to return. You can use '*' for
            all fields. A curated selection is used by default.
    Returns:
        (Series): The detailed patent data.
    References:
        - https://ped.uspto.gov/peds/#!/#%2FapiDocumentation
    """

    # Ensure that a patent number and patent URL are passed.
    if data is None and patent_number:
        patent = pd.Series({
            'patent_number': patent_number,
            'patent_url': patent_url,
        })
    elif data is not None:
        patent = data
        patent_number = patent['patent_number']
    else:
        raise ValueError

    # Specify the fields to return.
    if fields is None:
        fields = ['appType', 'appFilingDate', 'inventorName',
                  'inventors', 'patentIssueDate'] # 'attrnyAddr',
    if isinstance(fields, str):
        field_list = fields
    else:
        field_list = ','.join(fields)

    # TODO: Simply use DEFAULT_HEADERS if no headers provided.
        
    # Request fields for the patent.
    base = 'https://ped.uspto.gov/api/queries'
    headers = {'User-Agent': user_agent}
    data = {
        'searchText': f'{query}:({patent_number})',
        'fl': field_list, # field list
        'df': search_field, # default field to search
        'qf': search_fields, # multiple fields to search
        'facet': 'false',
        'mm': '100%', # minimum match
        'sort': f'{search_field} asc',
        'start': '0',
    }
    response = requests.post(base, json=data, headers=headers)

    # Add the patent details.
    data = response.json()
    docs = data['queryResults']['searchResponse']['response']['docs']
    doc = docs[0]
    doc = clean_dictionary(doc, function=camel_to_snake)

    # Optional: Get the attorney data.
    # doc['attorney'] = [clean_dictionary(x, function=camel_to_snake) for x in doc['attrny_addr']]
    # del doc['attrny_addr']
    
    # FIXME: Get the inventor data (from the text if not present here).
    try:
        doc['inventor_name'] = doc['inventor_name'].title()
        doc['inventors'] = [clean_dictionary(x, function=camel_to_snake) for x in doc['inventors']]
        inventor = doc['inventors'][0]
        doc['inventor_city'] = inventor['city'].replace(',', '').title()
        doc['inventor_state'] = inventor['geo_code']
        doc['inventor_country'] = inventor['country'].replace('(', '').replace(')', '')
    except KeyError:
        pass
    
    # Merge details with patent.
    patent = pd.concat([patent, pd.Series(doc)])

    # Get patent text by parsing the patent's webpage.
    response = requests.get(patent['patent_url'], headers=headers)
    soup = BeautifulSoup(response.content, features='html.parser')

    # Get the abstract.
    patent['abstract'] = soup.p.text.strip().replace('\n     ', ' ')

    # Get the applicant.
    tables = soup.findAll('table')
    values = [x.text.strip() for x in tables[3].findAll('td')]
    try:
        patent['applicant_name'] = values[2]
        patent['applicant_city'] = values[3]
        patent['applicant_state'] = values[4]
        patent['applicant_country'] = values[5]
    except IndexError:
        print('Error parsing applicant:', values)

    # Get the claims.
    # Test: Handle `It is claimed:`.
    claims = soup.text.split('claimed is:')[-1].split('claimed:')[0]
    claims = claims.split('Description  BACKGROUND OF THE INVENTION')[0]
    claims = claims.strip()
    claims = re.split('(\n\s\d.\s\s)', claims)
    claims = claims[::2]
    claims[0] = claims[0].replace('1.  ', '')
    patent['claims'] = [x.replace('\n', ' ').strip() for x in claims]

    # TODO: Download the PDF for the patent.
    # - Upload the PDF to Firebase Storage, saving the ref with the data.

    # Optional: Get plant details.

    # '\nSeeds\n '
    # Market Class: 
    # patent['parentage'] = soup.text.split('Parentage: ')[1].split(':')[0]
    # patent['classification'] = soup.text.split('Classification: ')[1].split(':')[0]

    # Optional: Get lab results?
    # soup.text.split('TABLE-US-')[1]

    # Optional: Get more patent details.
    # - citations_applicant_count
    # - full text (description)

    # Optional: Extract links to references.
    # tables[8].text

    # Bonus: Crop all images for each patent.
    # Upload the images to Firebase Storage for access online / through the API.

    # Return the augmented patent data.
    return patent


def get_strain_name(x):
    """Get a strain name in text surrounded by tildes."""
    try:
        return re.search('`(.*)`', x).group(1)
    except AttributeError:
        return ''


if __name__ == '__main__':

    #-------------------------------------------------------------------
    # 1. Find cannabis plant patents.
    # Future work: Download PDFs for the patents found.
    #-------------------------------------------------------------------

    # # Search for cannabis patents.
    queries = [
        # 'cannabis',
        # 'cannabis plant',
        'cannabis cultivar',
        # 'cannabis variety',
        # 'hemp plant',
        # 'hemp cultivar',
        # 'hemp variety',
        # 'marijuana plant',
    ]
    limit = 10
    for query in queries:

        # Search for patents by keyword(s).
        patents = search_patents(query, limit, term='TTL%2F')
        print('Found %i patents.' % len(patents))

        # Save the patents.
        key = kebab_case(query)
        datafile = f'../../.datasets/ai/plant-patents/{key}-patents.xlsx'
        to_excel_with_style(patents, datafile, sheet_name='Patents')

        # # Read the patents back in.
        # patents = pd.read_excel(datafile)

        # Isolate plant patents.
        # Note: There is probably a better way to identify plant patents.
        cultivars = patents.loc[
            (patents['patent_title'].str.contains('plant', case=False)) |
            (patents['patent_title'].str.contains('cultivar', case=False))
        ]
        print('Found %i cultivar patents.' % len(cultivars))

        # Get strain names.
        strain_names = cultivars['patent_title'].apply(get_strain_name)
        cultivars = cultivars.assign(strain_name=strain_names)

        # TODO: Remove duplicates: it appears that patents can be renewed.

        # Get details for each row.
        # Note: This could probably be done more efficiently.
        cultivar_data = []
        for _, cultivar in cultivars.iterrows():
            patent_details = get_patent_details(cultivar)
            cultivar_data.append(patent_details)
        cultivar_data = pd.concat(cultivar_data, axis=1)
        if isinstance(cultivar_data, pd.Series):
            cultivar_data = cultivar_data.to_frame()
        cultivar_data = cultivar_data.transpose()

        # Save the plant patent data.
        today = kebab_case(datetime.now().isoformat()[:16])
        datafile = f'../../.datasets/ai/plant-patents/plant-patents-{today}.xlsx'
        to_excel_with_style(cultivar_data, datafile, sheet_name='Patent Details')


    # # Lookup referenced cultivars:
    # # - Santhica 27
    # # - BLK03
    # # - AVI-1
    # patent = get_patent_details(pd.Series({
    #     'patent_number': 'PP34051',
    #     'patent_number_formatted': 'PP34,051',
    #     'patent_title': 'Cannabis plant named `AVI-1`',
    #     'patent_url': 'https://patft.uspto.gov/netacgi/nph-Parser?Sect1=PTO2&Sect2=HITOFF&u=%2Fnetahtml%2FPTO%2Fsearch-adv.htm&r=7&f=G&l=50&d=PTXT&p=1&S1=%22marijuana+plant%22&OS=%22marijuana+plant%22&RS=%22marijuana+plant%22',
    #     'strain_name': 'AVI-1',
        
    # }))
    # today = kebab_case(datetime.now().isoformat()[:16])
    # datafile = f'../../.datasets/ai/plant-patents/plant-patent-{today}.xlsx'
    # to_excel_with_style(patent.to_frame().T, datafile, sheet_name='Patent Details')


    #-------------------------------------------------------------------
    # 3. Organize all plant patent data points.
    # At this stage, a data guide can be written, with:
    # - Key
    # - Type
    # - Description
    # - Possible values
    # - Relation to other variables?
    #-------------------------------------------------------------------

    # Read manually collected plant patent data.
    datafile = '../../.datasets/ai/plant-patents/plant-patents.xlsx'
    results = pd.read_excel(datafile, sheet_name='Patent Lab Results')
    results['patent_number'] = results['patent_number'].astype(str)

    # Read programmatically collected plant patent data.
    datafile = '../../.datasets/ai/plant-patents/plant-patents.xlsx'
    details = pd.read_excel(datafile, sheet_name='Patent Details')
    details['patent_type'] = 'Plant'
    details['patent_link'] = 'http://www.pat2pdf.org/pat2pdf/foo.pl?number=' + details['patent_number'].astype(str)
    details['patent_number'] = details['patent_number'].astype(str)

    # Merge average lab results with details.
    avg_results = results.groupby('patent_number', as_index=False).mean()
    details = pd.merge(
        left=details,
        right=avg_results,
        left_on='patent_number',
        right_on='patent_number',
    )

    # Count plant patents over time.
    details['date'] = pd.to_datetime(details['patent_issue_date'])
    group = details.groupby(pd.Grouper(key='date', freq='Y'), as_index=True)
    yearly = group['patent_number'].count()
    yearly = yearly.to_frame()
    yearly['year'] = pd.to_datetime(yearly.index).strftime('%Y')
    yearly.rename(columns={'patent_number': 'plant_patents'}, inplace=True)
    annual_stats = yearly[['plant_patents', 'year']].to_dict(orient='records')

    # Optional: Calculate statistics from patent data.
    # - THCV to THC ratio.
    # - THCV to total cannabinoid ratio.
    # - THC / CBD ratio.
    # - CBD / CBC / THC ratio.
    # - All terpene ratios!!!

    # Ridge plots for cultivars with cannabinoid terpene data.
    # BONUS: Use as an image for the patent (chemotype fingerprint)!


    # Regression plots of ratios with all strains colored by strain.


    #-------------------------------------------------------------------
    # 4. Upload the patent data.
    #-------------------------------------------------------------------

    # Initialize Firebase!
    initialize_firebase(env_file='../../.env')

    # Update the base patents document with statistics.
    update_document('public/stats/patents/plant_patents', {
        'annual_plant_patents': annual_stats,
    })

    # Upload the plant patent data to Firestore for API access.
    col = 'public/data/patents'
    refs = [f'{col}/{x}' for x in details['patent_number']]
    data = details.to_dict(orient='records')
    update_documents(refs, data)

    # Upload the patent lab results data to Firestore for API access.
    results['id'] = results.index
    values = results[['patent_number', 'id']].values
    refs = [f'{col}/{x[0]}/patent_lab_results/lab_result_{x[1]}' for x in values]
    data = results.to_dict(orient='records')
    update_documents(refs, data)


    #-------------------------------------------------------------------
    # 5. TODO: Access the patent data through the Cannlytics API!
    #-------------------------------------------------------------------

    # import seaborn as sns

    # sample = results.groupby('strain_name').mean()
    # sns.regplot(
    #     x='d_limonene',
    #     y='beta_pinene',
    #     data=sample,
    #     ci=95,
    # )




# DEV:

from cannlytics.utils.constants import DEFAULT_HEADERS

# Selenium imports.
from selenium import webdriver
from selenium.webdriver.edge.options import Options as EdgeOptions
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait


# TODO: Turn into a re-usable function


# Initialize a web driver.
try:
    service = Service()
    options = Options()
    # options.add_argument('--window-size=1920,1200')
    # options.add_argument('--headless')
    # options.add_argument('--disable-gpu')
    # options.add_argument('--no-sandbox')
    driver = webdriver.Chrome(options=options, service=service)
except:
    options = EdgeOptions()
    # options.add_argument('--headless')
    driver = webdriver.Edge(options=options)

# Get the patent search page.
driver.get('https://ppubs.uspto.gov/pubwebapp/static/pages/ppubsbasic.html')

# Search for patents.
search_terms = {
    'searchText1': 'cannabis',
    'searchText2': 'cultivar'
}
for element_id, term in search_terms.items():
    input_element = driver.find_element(By.ID, element_id)
    input_element.clear()
    input_element.send_keys(term)

# Click the search button.
search_button = driver.find_element(by=By.ID, value='searchText2')
search_button.send_keys(Keys.ENTER)
sleep(3)

# Iterate over search result pages.
iterate = True
patents = []
while iterate:

    # Wait until the overlay is no longer visible.
    WebDriverWait(driver, 3).until(EC.invisibility_of_element((By.CSS_SELECTOR, '.overlay.full-page')))

    # Extract the page number and total number of pages.
    page_info_span = driver.find_element(By.ID, 'pageInfo')
    page_info_text = page_info_span.text
    current_page, total_pages = page_info_text.split(" of ")
    current_page = current_page.replace('Page ', '')
    
    # Get the metadata for each patent.
    metadata = []
    table = driver.find_element(by=By.ID, value='searchResults')
    rows = table.find_elements(by=By.TAG_NAME, value='tr')[1:]
    print('Getting %i patents on page' % len(rows), current_page, 'of', total_pages)
    for row in rows:
        metadata.append({
            'query_number': row.find_element(By.XPATH, './td[1]').text,
            'patent_number': row.find_element(By.XPATH, './td[2]').text,
            'patent_title': row.find_element(By.XPATH, './td[4]').text,
            'inventor_name': row.find_element(By.XPATH, './td[5]').text,
            'date_published': row.find_element(By.XPATH, './td[6]').text,
            'page_count': int(row.find_element(By.XPATH, './td[7]').text)
        })

    # Add URLs to the list of dictionaries.
    count = 0
    links = driver.find_elements(by=By.TAG_NAME, value='a')
    for i, link in enumerate(links):
        href = link.get_attribute('href')
        if href and 'image-ppubs.uspto.gov' in href:
            metadata[count]['patent_url'] = href
            count += 1
    
    # Record the patent data.
    patents.extend(metadata)
    
    # Click the next button if it's not the last page.
    if current_page != total_pages:
        next_button = WebDriverWait(driver, 3).until(
            EC.element_to_be_clickable((By.ID, 'paginationNextItem'))
        )
        next_button.click()
    else:
        iterate = False

# Close the driver.
print('Found %i patents, closing driver.' % len(patents))
driver.quit()

# Save the patent metadata.
print('Saving patent metadata...')
date = datetime.now().strftime('%Y-%m-%d-%H-%M-%S')
metadata_file = f'D://data/strains/patents/patent-metadata-{date}.xlsx'
patent_data = pd.DataFrame(patents)
patent_data.to_excel(metadata_file)
print('Patent metadata saved:', metadata_file)

# Download patent images.
print('Downloading patent images.')
pdf_dir = 'D://data/strains/patents/pdfs/'
for patent in patents:
    outfile = os.path.join(pdf_dir, patent['patent_number'] + '.pdf')
    if os.path.exists(outfile):
        print('Cached:', outfile)
        continue
    response = requests.get(patent['patent_url'])
    with open(outfile, 'wb') as file:
        file.write(response.content)
    print('Downloaded:', outfile)


#------------------------------------------------------------
# Use re-usable patent search function.
#------------------------------------------------------------

# TODO: Iterate over search terms.
queries = [
    ('cannabis', ''),
    ('cannabis', 'plant'),
    ('cannabis', 'cultivar'),
    ('cannabis', 'variety'),
    ('hemp', ''),
    ('hemp', 'plant'),
    ('hemp', 'cultivar'),
    ('hemp', 'variety'),
    ('marijuana', ''),
    ('marijuana', 'plant'),
    ('marijuana', 'cultivar'),
    ('marijuana', 'variety'),
]


#-----------------------------------------------------------------------
# Get patents with strain names in their titles.
#-----------------------------------------------------------------------

# Isolate plant patents.
strains = patent_data.loc[
    (patent_data['patent_title'].str.contains('plant', case=False)) |
    (patent_data['patent_title'].str.contains('cultivar', case=False))
]
print('Found %i cultivar patents.' % len(strains))

# Extract all text after "named"
strain_names = strains['patent_title'].str.extract(r"named (.*)")

# Remove apostrophes from the extracted strain names
strain_names = strain_names.replace("'", "")

# Remove non-standard characters from the strain names
strains = strains.assign(strain_name=strain_names)
strains['strain_name'] = strains['strain_name'].apply(
    lambda x: re.sub(r'[^a-zA-Z0-9\s]', '', str(x)) \
        if pd.notnull(x) else x).str.strip()

# Keep only strains with strain names.
strains = strains.loc[strains['strain_name'].notnull()]
print(strains['strain_name'].unique())


#------------------------------------------------------------
# Extract data from downloaded patents.
#------------------------------------------------------------

import os
from PIL import Image

import tempfile
from pdf2image import convert_from_path
import pdfplumber
from pytesseract import image_to_pdf_or_hocr, image_to_string

# output_folder = r'D:\data\strains\patents\images'
# original_file = r'D:\data\strains\patents\pdfs\US-5589367-A.pdf'
# pdf_image = r'D:\data\strains\patents\images\US-5589367-A.png'

# Define the output directories.
pdf_dir = 'D://data/strains/patents/pdfs/'
image_dir = 'D://data/strains/patents/images/'

# Extract the text from all strain patents.
extracted_data = {}
for index, strain in strains.iterrows():
    patent_number = strain['patent_number']
    original_file = os.path.join(pdf_dir, patent_number + '.pdf')
    pdf_image = os.path.join(image_dir, patent_number + '.png')

    # Check the number of pages.
    report = pdfplumber.open(original_file)
    page_count = len(report.pages)
    print('Parsing patent %s (%i pages)...' % (patent_number, page_count))

    # Save the first page as an image.
    im = report.pages[0].to_image(resolution=90)
    im.save(pdf_image, format='PNG')

    # Close the report.
    report.close()

    # Read the text of the first page.
    page_text = image_to_string(pdf_image)
    abstract = page_text.split('ABSTRACT')[1].split('Claims')[0]

    # Record the extracted data.
    obs = {'abstract': abstract}
    extracted_data[patent_number] = {**strain.to_dict(), **obs}
    print('Extracted data for patent:', patent_number)


# Explore the abstracts with OpenAi's ChatGPT model!
for patent_number, obs in extracted_data.items():
    print(len(obs['abstract'].split(' ')))

    # TODO: Use the tokenizer to predict cost to parse.


# TODO: Save parsed PDFs and make them available though Cannlytics!


# pdf = image_to_pdf_or_hocr(original_file, extension='pdf')

# Read the text of the first page.
# texts = []
# images = convert_from_path(
#     original_file,
#     output_folder=output_folder,
#     dpi=90,
# )
# for image in images[:1]:
#     texts.append(image_to_string(image))

    # images[i].save('image_name'+ str(i) +'.jpg', 'JPEG')
# with tempfile.TemporaryDirectory() as path:
#     images = convert_from_path(original_file, output_folder=path)
#     for img in images[:1]:
#         texts.append(image_to_string(img))


# doc = convert_from_path(original_file)
# path, fileName = os.path.split(original_file)
# fileBaseName, fileExtension = os.path.splitext(fileName)

# for page_number, page_data in enumerate(doc):
#     txt = image_to_string(Image.fromarray(page_data)).encode('utf-8')
#     print("Page # {} - {}".format(str(page_number),txt))

# # Convert PDF to images.
# images = convert_from_path(pdf_path)
# text = []

# # Apply OCR to each image
# for i, img in enumerate(images):
#     text.append(image_to_string(img))

# return text


#-----------------------------------------------------------------------
# TODO: Apply OCR to downloaded images.
#-----------------------------------------------------------------------

# def ocr_pdf(pdf_path):
#     # Convert PDF to images
#     images = convert_from_path(pdf_path)
#     text = []
    
#     # Apply OCR to each image
#     for i, img in enumerate(images):
#         text.append(pytesseract.image_to_string(img))
    
#     return text

#-----------------------------------------------------------------------
# TODO: Extract data from generated PDFs.
#-----------------------------------------------------------------------

# def extract_data_from_pdfs(pdf_dir):
#     for filename in os.listdir(pdf_dir):
#         if filename.endswith(".pdf"):
#             pdf_path = os.path.join(pdf_dir, filename)
            
#             # Get metadata
#             metadata = get_patent_metadata(pdf_path)
#             print(f"Metadata for {filename}:", metadata)
            
#             # Apply OCR and print text
#             text = ocr_pdf(pdf_path)
#             print(f"Text for {filename}:", text)


# Data points sought:
# - patent_number_formatted
# - patent_type
# - patent_link
# - patent_issue_date
# - patent_year
# - patent_month
# - app_type
# - app_filing_date
# - inventors
# - inventor_name
# - inventor_city
# - inventor_state
# - strain_name
# - key
# - type
# - description
# - lineage
# - average total cannabinoids
# - average total thc
# - average total cbd
# - average total terpenes
# - major terpenes
# - THCV to THC ratio.
# - THCV to total cannabinoid ratio.
# - THC / CBD ratio.
# - CBD / CBC / THC ratio.
# - All terpene ratios!!!


# TODO: Save the extracted patent data.



"""
Calculation Parser | Cannlytics
Copyright (c) 2021-2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 6/8/2021
Updated: 6/8/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>
"""
# Standard imports
import re

# External imports
try:
    import pandas as pd
    import numpy as np
except:
    pass


def calculate_results():
    """Calculate results by using analyte formula and
    instrument / analyst measurements.
    Calculate results by using analyte formula and
    instrument / analyst measurements.
    """
    # TODO: Implement a function that calculates results given a formula and results.
    raise NotImplementedError


# # Prepare testing dataset
# tags = np.array(['tag'+str(i) for i in np.random.randint(10, size=200)])  # randomly generate tag list
# vals = np.random.randint(20, size=200)  # generate a list of random integers

# raw_df = pd.DataFrame({
#     'tag': tags,
#     'value': vals
# })

# # Functions
# def parentheses_enclosed(s):
#     paren_order = re.findall(r'[\(\)]', s)
    
#     if paren_order.count('(') != paren_order.count(')'):
#         return False
    
#     curr_levels = []
#     nest_lv = 0
#     for p in paren_order:
#         if p == '(':
#             nest_lv += 1
#         else:
#             nest_lv -= 1
#         curr_levels.append(nest_lv)
#     if 0 in curr_levels[:-1]:
#         return False
#     else:
#         return True
    
    
# def remove_matched_parentheses(s):
#     if ')' in s:
#         # find the first ')'
#         end = s.find(')')
#         # find the last '(' before the first ')'
#         start = max([i for i, char in enumerate(s[:end]) if char == '(' ])
#         # remove the parentheses
#         return remove_matched_parentheses(s[:start] + s[end+1:])
#     else:
#         return s
    

# def interpret(f, df):
#     if re.match(r'\Atag[\d]+\Z', f):  # e.g. 'tag1'
#         return df[df.tag == f]['value'].values
    
#     elif parentheses_enclosed(f) and \
#         re.match(r'\Asum\(.+[\+\-].+\)\Z|\Aavg\(.+[\+\-].+\)\Z|\Amin\(.+[\+\-].+\)\Z|\Amax\(.+[\+\-].+\)\Z', f):
        
#         f_name = f[:3]  # get agg func name
#         f_stripped = f[4:-1]  # strip outer func
        
#         while re.match(r'\A\(.+\)\Z', f_stripped) and parentheses_enclosed(f_stripped):
#             f_stripped = f_stripped[1:-1]
            
#         comps = re.compile(r'[\+\-]').split(f_stripped)  # split by + or -
        
#         operators = re.findall(r'[\+\-]', f_stripped)
#         comps_final = []
#         temp_str = ''
#         for c in comps:
#             temp_str += c
#             if re.match(r'\Atag[\d]+\Z', temp_str) or parentheses_enclosed(temp_str):
#                 comps_final.append(f'{f_name}({temp_str})')
#                 if len(operators) > 0:
#                     comps_final.append(operators.pop(0))
#                 temp_str = ''
#             else:
#                 temp_str += operators.pop(0)
#         return interpret(''.join(comps_final), df)
    
#     elif re.match(r'\Asum\([^\(\)]+\)\Z', f):  # e.g. 'sum(tag1)'
#         return np.sum(interpret(f[4:-1], df))
    
#     elif re.match(r'\Aavg\([^\(\)]+\)\Z', f):  # e.g. 'avg(tag1)'
#         return np.average(interpret(f[4:-1], df))
    
#     elif re.match(r'\Amin\([^\(\)]+\)\Z', f):  # e.g. 'min(tag1)'
#         return np.min(interpret(f[4:-1], df))
    
#     elif re.match(r'\Amax\([^\(\)]+\)\Z', f):  # e.g. 'max(tag1)'
#         return np.max(interpret(f[4:-1], df))
    
#     elif re.match(r'\A\(.+\)\Z', f) and parentheses_enclosed(f):  # e.g. '(tag1-tag2)'
#         return interpret(f[1:-1], df)
    
#     elif f.replace('.', '', 1).isdigit():
#         return float(f)
    
#     else:
#         rest_f = remove_matched_parentheses(f)
#         if '+' in rest_f or '-' in rest_f:
#             comps = re.compile(r'[\+\-]').split(f)
#         else:
#             comps = re.compile(r'[\*\/]').split(f)
            
#         if comps[0].count('(') != comps[0].count(')'):
#             nested_level = comps[0].count('(') - comps[0].count(')')
#             pos = len(comps[0])
#             for comp in comps[1:]:
#                 if '(' in comp:
#                     nested_level += comp.count('(')
#                 if ')' in comp:
#                     nested_level -= comp.count(')')
#                 pos += len(comp) + 1  # +1 because of the operator inside parenthesis
#                 if nested_level == 0:
#                     break
#         else:
#             pos = len(comps[0])
        
#         left = f[:pos]  # left component
#         right = f[pos+1:]  # right component
#         operator = f[pos]  # the operator
#         if operator == '+':
#             return interpret(left, df) + interpret(right, df)
#         elif operator == '-':
#             return interpret(left, df) - interpret(right, df)
#         elif operator == '*':
#             return interpret(left, df) * interpret(right, df)
#         elif operator == '/':
#             denominator = interpret(right, df)
#             if denominator == 0 or denominator is np.nan: 
#                 return np.nan
#             else:
#                 return interpret(left, df) / interpret(right, df)
    
#     return np.nan
    

# # Verification
# assert np.sum(raw_df[raw_df.tag == 'tag1']['value'].values) == interpret('sum(tag1)', raw_df), "Wrong!"
# assert np.average(raw_df[raw_df.tag == 'tag1']['value'].values) == interpret('avg(tag1)', raw_df), "Wrong!"
# assert np.min(raw_df[raw_df.tag == 'tag1']['value'].values) == interpret('min(tag1)', raw_df), "Wrong!"
# assert np.max(raw_df[raw_df.tag == 'tag1']['value'].values) == interpret('max(tag1)', raw_df), "Wrong!"

# assert np.sum(raw_df[raw_df.tag == 'tag1']['value'].values) + \
#     np.sum(raw_df[raw_df.tag == 'tag2']['value'].values) == \
#     interpret('sum(tag1)+sum(tag2)', raw_df), "Wrong!"

# assert np.sum(raw_df[raw_df.tag == 'tag1']['value'].values) + \
#     np.sum(raw_df[raw_df.tag == 'tag2']['value'].values) + \
#     np.average(raw_df[raw_df.tag == 'tag3']['value'].values) == \
#     interpret('sum(tag1)+sum(tag2)+avg(tag3)', raw_df), "Wrong!"

# assert np.sum(raw_df[raw_df.tag == 'tag1']['value'].values) + \
#     np.sum(raw_df[raw_df.tag == 'tag2']['value'].values) * \
#     np.average(raw_df[raw_df.tag == 'tag3']['value'].values) == \
#     interpret('sum(tag1)+sum(tag2)*avg(tag3)', raw_df), "Wrong!"

# assert (
#     np.sum(raw_df[raw_df.tag == 'tag1']['value'].values) + \
#     (np.sum(raw_df[raw_df.tag == 'tag2']['value'].values) - \
#     np.sum(raw_df[raw_df.tag == 'tag3']['value'].values)) + 10
# ) * (
#     np.max(raw_df[raw_df.tag == 'tag4']['value'].values) + \
#     np.average(raw_df[raw_df.tag == 'tag5']['value'].values)
# ) * 0.2 == interpret('(sum(tag1+(tag2-tag3))+10)*(max(tag4)+avg(tag5))*0.2', raw_df), "Wrong!"

# print('All pass!')



"""
Report Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Report data model.
"""
# Standard imports.
from dataclasses import dataclass
from datetime import datetime

# Internal imports.
from .base import Model

@dataclass
class Report(Model):
    """."""
    _collection = 'organizations/%s/reports'



"""
Cannlytics Module Initialization | Cannlytics
Copyright (c) 2021-2023 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/6/2021
Updated: 7/1/2023
"""
# from .cannlytics import Cannlytics
import cannlytics.auth as auth
import cannlytics.data as data
import cannlytics.firebase as firebase
import cannlytics.lims as lims
import cannlytics.metrc as metrc
import cannlytics.models as models
import cannlytics.stats as stats
import cannlytics.utils as utils


__all__ = [
    # 'Cannlytics',
    'auth',
    'data',
    'firebase',
    'lims',
    'metrc',
    'models',
    'stats',
    'utils',
]
__title__ = 'cannlytics'
__version__ = '0.0.15'
__author__ = 'Keegan Skeate <https://github.com/keeganskeate>'
__license__ = 'MIT <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>'
__copyright__ = 'Copyright (c) 2022-2023 Cannlytics'



"""
Instruments | Cannlytics
Copyright (c) 2021-2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 8/3/2021
Updated: 12/21/2021
License: MIT License <https://opensource.org/licenses/MIT>

Description:
    
    Manage scientific instruments and measurements from the instruments.

"""
# Standard imports
import os
from datetime import datetime, timedelta
from typing import Any, List, Optional

# External imports
from dotenv import dotenv_values
import pandas as pd
import requests

# Internal imports
from ..firebase import (
    get_collection,
    initialize_firebase,
    update_document
)
from ..utils.utils import clean_column_names, snake_case


API_BASE = 'https://console.cannlytics.com'


# TODO: Refactor, big time.
def automatic_collection(
        org_id: Optional[str] = None,
        env_file: Optional[str] = '.env',
        minutes_ago: Optional[int] = None,
) -> List[dict]:
    """Automatically collect results from scientific instruments.
    Args:
        org_id (str): The organization ID to associate with instrument results.
        env_file (str): The environment variable file, `.env` by default.
            Either a `GOOGLE_APPLICATION_CREDENTIALS` or a
            `CANNLYTICS_API_KEY` is needed to run the routine.
        minutes_ago (int): The number of minutes in the past to restrict
            recently modified files.
    Returns:
        (list): A list of measurements (dict) that were collected.
    """

    # Try to set credentials and initialize Firebase, otherwise an API key will be used.
    try:
        config = dotenv_values(env_file)
        credentials = config.get('GOOGLE_APPLICATION_CREDENTIALS')
        initialize_firebase(credentials)
    except:
        pass

    # Get the organization ID from the .env file if not specified.
    if not org_id:
        org_id = config['CANNLYTICS_ORGANIZATION_ID']

    # Format the last modified time cut-off as a datetime.
    last_modified_at = None
    if minutes_ago:
        last_modified_at = datetime.now() - timedelta(minutes=minutes_ago)

    # Get the instruments, trying Firestore, then the API.
    try:
        ref = f'organizations/{org_id}/instruments'
        instrument_data = get_collection(ref)
    except:
        api_key = config['CANNLYTICS_API_KEY']
        headers = {
            'Authorization': 'Bearer %s' % api_key,
            'Content-type': 'application/json',
        }
        url = f'{API_BASE}/instruments?organization_id={org_id}'
        response = requests.get(url, headers=headers)
        instrument_data = response.json()['data']

    # TODO: Split this functionality into separate functions.
    # Iterate over instruments, collecting measurements.
    measurements = []
    for instrument in instrument_data:

        # Iterate over analyses that the instrument may be running.
        try:
            analyses = instrument.get('analyses', '').split(',')
        except AttributeError:
            continue # FIXME: Handle missing analyses more elegantly.
        analyses = [x.strip() for x in analyses]
        for index, analysis in enumerate(analyses):

            # Optional: Handle multiple data paths more elegantly.
            analysis = analyses[index]
            try:
                data_paths = instrument['data_path'].split(',')
            except AttributeError:
                continue # No data path.
            data_paths = [x.strip() for x in data_paths]
            data_path = data_paths[index]
            if not data_path:
                continue

            # Identify the analysis being run and identify the import routine.
            # Optional: Identify more elegantly.
            if 'micro' in analysis or 'MICR' in analysis:
                import_method = globals()['import_micro']
            elif 'metal' in analysis or 'HEAV' in analysis:
                import_method = globals()['import_heavy_metals']
            else:
                import_method = globals()['import_results']

            # Search for recently modified files in the instrument directory
            # and parse any recently modified file.
            for root, _, filenames in os.walk(data_path):
                for filename in filenames:
                    if filename.endswith('.xlsx') or filename.endswith('.xls'):
                        data_file = os.path.join(root, filename)
                        modifed_at = os.stat(data_file).st_mtime
                        # TODO: Ensure that date restriction works.
                        if last_modified_at:
                            if modifed_at < last_modified_at:
                                continue
                        samples = import_method(data_file)
                        if isinstance(samples, dict):
                            sample_data = {**instrument, **samples}
                            measurements.append(sample_data)
                        else:
                            for sample in samples:
                                sample_data = {**instrument, **sample}
                                measurements.append(sample_data)

    # Upload measurement data to Firestore.
    now = datetime.now()
    updated_at = now.isoformat()
    for measurement in measurements:
        try:
            measurement['sample_id'] = measurement['sample_name']
        except:
            continue # Already has `sample_id`.

        # TODO: Format a better measurement ID.
        measurement_id = measurement.get('acq_inj_time') # E.g. 12-Jun-21, 15:21:07
        if not measurement_id:
            measurement_id = measurement['sample_id']
        else:
            try:
                measurement_id = measurement_id.replace(',', '').replace(' ', '-').replace(':', '-')
            except AttributeError:
                pass
            measurement_id = str(measurement_id) + '_' + str(measurement['sample_id'])
        measurement['measurement_id'] = measurement_id
        measurement['updated_at'] = updated_at
        ref = f'organizations/{org_id}/measurements/{measurement_id}'
        try:
            update_document(ref, measurement)
        except:
            url = f'{API_BASE}/measurements/{measurement_id}?organization_id={org_id}'
            response = requests.post(url, json=measurement, headers=headers)
        print('Uploaded measurement:', ref)

        # Upload result data to Firestore
        for result in measurement['results']:
            analyte = result['analyte']
            result_id = f'{measurement_id}_{analyte}'
            result['sample_id'] = measurement['sample_name']
            result['result_id'] = result_id
            result['measurement_id'] = measurement_id
            result['updated_at'] = updated_at
            ref = f'organizations/{org_id}/results/{result_id}'
            try:
                update_document(ref, result)
            except:
                url = f'{API_BASE}/results/{result_id}?organization_id={org_id}'
                response = requests.post(url, json=result, headers=headers)
            print('Uploaded result:', ref)

    # Return the measurements
    return measurements


def get_compound_dataframe(workbook_data: dict, sheetname: Optional[str] = 'Compound') -> Any:
    """ Rename the columns in the `Compound` sheet to match the required
    names. For simplicity, make a copy of the `Compound` sheet to
    handle NaN values.
    Args:
        workbook_data (dict): A dictionary of worksheet DataFrames.
        sheetname (str): The worksheet that contains the compound data.
    Returns:
        (DataFrame): A DataFrame with renamed compounds.
    """
    columns = {
        'Name':'analyte',
        'Amount': 'measurement',
        'AmtPerResp': 'amount_per_response'
    }
    workbook_data[sheetname].rename(columns=columns, inplace=True)
    compounds = workbook_data[sheetname].copy()
    criterion = (compounds.analyte.isnull()) & (compounds.measurement > 0)
    compounds.loc[criterion, 'analyte'] = 'wildcard'
    compounds.dropna(subset=['analyte'], inplace=True)
    return compounds


def get_sample_data(workbook_data: dict, sheet_name: Optional[str] = 'Sheet1') -> dict:
    """Return the sample name from a dictionary.
    Converts the first column of the first sheet to snake_case.
    Args:
        data (DataFrame): A DataFrame.
        sheet_name (str): The name of the worksheet containing the sample data.
    Returns:
        (dict): The sample data as a key and value pairs.
    """
    data = workbook_data[sheet_name]
    data['ObjClass'] = data['ObjClass'].apply(snake_case)
    return dict(data.values)


def import_analyses(directory: str):
    """Import analyses to Firestore from a .csv or .xlsx file.
    Args:
        directory (str): The full filename of a data file.
    """
    analyses = pd.read_excel(directory + 'analyses.xlsx')
    analytes = pd.read_excel(directory + 'analytes.xlsx')
    for index, analysis in analyses.iterrows():
        analyte_data = []
        analyte_names = analysis.analyte_keys.split(', ')
        for analyte_key in analyte_names:
            analyte_item = analytes.loc[analytes.key == analyte_key]
            analyte_data.append(analyte_item.to_dict(orient='records'))
        analyses.at[index, 'analytes'] = analyte_data 
    analyses_data = analyses.to_dict(orient='records')
    # TODO: Implement.
    raise NotImplementedError


def import_heavy_metals(file_name: str) -> List[dict]:
    """Import heavy metal results from ICP-MS screening.
    First, reads in the log sheet and summary sheets seperately to
    make parsing easier. Drops the rows that do not contain sample ids.
    Renames columns to make parsing clearer. Get list of samples and
    then parses measurements.
    Args:
        data (DataFrame): A DataFrame.
    Returns:
        (list): A list of measurements (dict).
    """
    results_data = pd.read_excel(file_name, sheet_name='Log')
    samples_data = pd.read_excel(file_name, sheet_name='Quant Summary')
    results_data.dropna(subset = ['Sample Mass (g)'], inplace=True)
    samples_data.rename(columns = {'Analysis':'analyte', '-': 'mass'}, inplace=True)
    sample_ids = results_data['Sample ID'].tolist()
    samples = []
    for sample_id in sample_ids:
        sample = {}
        result_data = results_data[results_data['Sample ID'] == sample_id]
        sample['sample_id'] = sample_id
        sample['sample_mass'] = result_data['Sample Mass (g)'].values[0]
        sample['sample_dilution'] = result_data['Sample Dilution'].values[0]
        analytes = []
        criterion = (samples_data['analyte'] == 'ID:') & (samples_data['mass'] == sample_id)
        index = samples_data.index[criterion].tolist()
        for offset in range(index[0] + 20, index[0] + 27): # Magic numbers
            analyte = {}
            analyte['analyte'] = samples_data.iloc[offset].analyte
            analyte['measurement'] = samples_data.iloc[offset + 9].mass # Magic number
            analytes.append(analyte)
        sample['results'] = analytes
        samples.append(sample)
    return samples


def import_micro(file_name: str) -> List[dict]:
    """Import microbiological screening results, grouping sample results
    by well name.
    Args:
        file_name (str): The path to a qPCR data file.
    Returns:
        (list): A list of sample measurements (dict).
    """
    workbook_data = pd.read_excel(file_name, sheet_name=None)
    results = workbook_data['Tabular Results']
    sample_names = results['Well Name'].unique()
    columns = {
        'Dye': 'method',
        'Target': 'analyte',
        'Cq (∆R)': 'measurement',
        'Well': 'well',
        'Well Name': 'well_name',
    }
    results.rename(columns=columns, inplace=True)
    samples = []
    for sample_name in sample_names:
        sample = {'sample_name': sample_name}
        sample_data = results.loc[results['well_name'] == sample_name]
        fields = list(columns.values())
        sample['results'] = sample_data[fields].to_dict('records')
        samples.append(sample)
    return samples


def import_results(file_name: str) -> dict:
    """Import scientific instrument data. The routine is as follows:
        1. Read in all the excel sheets at one time.
        2. Get the sample name.
        3. Get and tidy the compound data from the compound sheet.
        4. Add the analyte names and measurements to a list of result dictionaries.
        5. Collect the data into a running list.
    Args:
        data (DataFrame): A DataFrame.
    Returns:
        (dict): A dictionary of sample results.
    """
    workbook_data = pd.read_excel(file_name, sheet_name=None)
    sample = get_sample_data(workbook_data)
    compounds = get_compound_dataframe(workbook_data)
    compounds = clean_column_names(compounds, 'analyte')
    columns = ['analyte', 'measurement', 'amount_per_response']
    sample['results'] = compounds[columns].to_dict('records')
    return sample


if __name__ == '__main__':

    # Standard imports.
    import argparse
    import os

    # Internal imports.
    from ..firebase import (
        get_collection,
        initialize_firebase,
        update_document
    )
    from ..utils.utils import snake_case

    # Declare command line arguments.
    parser = argparse.ArgumentParser(description='Scientific instrument data collection routine.')
    parser.add_argument('--env', action='store', dest='env_file', default='.env')
    parser.add_argument('--modified', action='store', dest='last_modified', default=None)
    parser.add_argument('--org', action='store', dest='org_id', default='')
    args = parser.parse_args()

    # Collect data.
    automatic_collection(args.org_id, env_file=args.env_file, minutes_ago=args.last_modified)



"""
Exceptions | Cannlytics
Copyright (c) 2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 7/11/2022
Updated: 7/11/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Invoice data model.
"""


class CannlyticsError(Exception):
    """A base class for Cannlytics system exceptions."""


class CannlyticsAPIError(CannlyticsError):
    """A primary error raised by the Cannlytics API."""

    def __init__(self, response):
        message = self.get_error_message(response)
        super().__init__(message)
        self.response = response


    def get_error_message(self, response):
        """Extract error message from a Cannlytics API response.
        Args:
            response (Response): A request response from the Cannlytics API.
        Returns:
            (str): Returns any error messages.
        """
        try:
            errors = response.json()
            if isinstance(errors, list):
                try:
                    message = '\n'.join(errors)
                except TypeError:
                    message = '\n'.join([x.get('message') for x in errors])
            elif isinstance(errors, dict):
                message = errors.get('message')
            else:
                message = response.text
        except (AttributeError, KeyError, ValueError):
            message = 'Unknown Cannlytics API error'
        return message



"""
Analyte Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Analyte data model.
"""
# Standard imports.
from dataclasses import dataclass
from typing import Optional

# Internal imports.
from .base import Model

@dataclass
class Analyte(Model):
    """An analyte is a specific measurement, such as the concentration
    of THCA, CBDV, or piperonyl butoxide. An analyte can have many
    supporting fields, such as lowest order of detection (LOD),
    lowest order of quantification (LOQ), regulatory limit and more."""
    _collection = 'organizations/%s/analytes'
    cas: str = ''
    public: bool = False
    formula: str = ''
    name: str = ''
    key: str = ''
    limit: Optional[float] = None
    lod: Optional[float] = None
    loq: Optional[float] = None
    units: Optional[str] = ''
    calculation_id: str = ''



"""
Parse Confident Cannabis CoA
Copyright (c) 2022 Cannlytics

Authors:
    Keegan Skeate <https://github.com/keeganskeate>
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
Created: 7/15/2022
Updated: 8/30/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse a Confident Cannabis CoA PDF or URL. Labs include:

        - CaliGreen Laboratory
        - California Ag Labs
        - CB Labs Novato
        - Harrens Lab Inc

Data Points:

    ✓ analyses
    - {analysis}_method
    ✓ {analysis}_status
    ✓ classification
    ✓ coa_urls
    ✓ date_tested
    - date_received
    ✓ images
    ✓ lab_results_url
    ✓ producer
    ✓ product_name
    ✓ product_type
    ✓ predicted_aromas
    ✓ results
    - sample_weight
    - total_cannabinoids (calculated)
    ✓ total_thc
    ✓ total_cbd
    - total_terpenes (calculated)
    ✓ sample_id (generated)
    ✓ strain_name
    ✓ lab_id
    ✓ lab
    ✓ lab_image_url
    - lab_license_number
    ✓ lab_address
    ✓ lab_city
    - lab_county (augmented)
    ✓ lab_state
    ✓ lab_zipcode
    ✓ lab_phone
    ✓ lab_email
    - lab_latitude (augmented)
    - lab_longitude (augmented)

"""
# Standard imports.
from datetime import datetime
import json
from time import sleep
from typing import Any, Optional

# External imports.
import pandas as pd
from pdfplumber.pdf import PDF
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.common.by import By
from selenium.webdriver.chrome.service import Service
from selenium.common.exceptions import (
    ElementNotInteractableException,
    NoSuchElementException,
    TimeoutException,
)
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait
try:
    import chromedriver_binary  # Adds chromedriver binary to path.
except ImportError:
    pass # Otherwise, ChromeDriver should be in your path.

# Internal imports.
from cannlytics import __version__
from cannlytics.data.data import create_hash, create_sample_id
from cannlytics.utils.utils import (
    convert_to_numeric,
    snake_case,
    strip_whitespace,
)


# It is assumed that the lab has the following details.
CONFIDENT_CANNABIS = {
    'coa_algorithm': 'confidentcannabis.py',
    'coa_algorithm_entry_point': 'parse_cc_coa',
    'lims': 'Con\x00dent Cannabis',
    'url': 'https://orders.confidentcannabis.com',
    'public': True,
}


def parse_cc_url(
        parser,
        url: str,
        headers: Optional[Any] = None,
        max_delay: Optional[float] = 60,
        persist: Optional[bool] = False,
        **kwargs
    ) -> dict:
    """Parse a Confident Cannabis CoA URL.
    Args:
        url (str): The CoA URL.
        headers (Any): Optional headers for standardization.
        max_delay (float): The maximum number of seconds to wait
            for the page to load.
        persist (bool): Whether to persist the driver.
            The default is `False`. If you do persist
            the driver, then make sure to call `quit`
            when you are finished.
    Returns:
        (dict): The sample data.
    """

    # Load the lab results with Selenium.
    if parser.service is None:
        parser.service = Service()
        parser.options = Options()
        parser.options.add_argument('--window-size=1920,1200')

        # DEV: Uncomment for development / comment for production.
        # parser.options.headless = False

        # PRODUCTION: Uncomment for production / comment for development.
        parser.options.add_argument('--headless')
        parser.options.add_argument('--disable-gpu')
        parser.options.add_argument('--no-sandbox')

    # Create a driver.
    if parser.driver is None:
        parser.driver = webdriver.Chrome(
            options=parser.options,
            service=parser.service,
        )

    # Get the URL.
    parser.driver.get(url)

    # Wait for the page to load by waiting to detect the image.
    try:
        el = (By.CLASS_NAME, 'product-box-cc')
        detect = EC.presence_of_element_located(el)
        WebDriverWait(parser.driver, max_delay).until(detect)
    except TimeoutException:
        print('Failed to load page within %i seconds.' % max_delay)

    # Create a sample observation.
    analyses, results = [], []
    obs = {'lims': 'Confident Cannabis'}

    # Find the sample image.
    el = parser.driver.find_element(
        by=By.CLASS_NAME,
        value='product-box-cc'
    )
    img = el.find_element(by=By.TAG_NAME, value='img')
    image_url = img.get_attribute('src')
    filename = image_url.split('/')[-1]
    obs['images'] = [{'url': image_url, 'filename': filename}]

    # Try to get the sample details.
    el = parser.driver.find_element(
        by=By.CLASS_NAME,
        value='product-desc',
    )
    block = el.text.split('\n')
    product_name = block[0]
    strain_name, product_type = tuple(block[3].split(', '))
    obs['product_name'] = product_name
    obs['lab_id'] = block[1]
    obs['classification'] = block[2]
    obs['strain_name'] = strip_whitespace(strain_name)
    obs['product_type'] = strip_whitespace(product_type)

    # Get the date tested.
    el = parser.driver.find_element(by=By.CLASS_NAME, value='report')
    span = el.find_element(by=By.TAG_NAME, value='span')
    tooltip = span.get_attribute('uib-tooltip')
    tested_at = tooltip.split(': ')[-1]
    date_tested = pd.to_datetime(tested_at).isoformat()
    obs['date_tested'] = date_tested

    # Get the CoA URL.
    button = el.find_element(by=By.TAG_NAME, value='button')
    href = button.get_attribute('href')
    base = url.split('/report')[0]
    coa_url = base.replace('/#!', '') + href
    filename = url.split('/')[-1].split('?')[0] + '.pdf'
    obs['coa_urls'] = [{'url': coa_url, 'filename': filename}]

    # Find the `analyses` and `results`.
    els = parser.driver.find_elements(by=By.CLASS_NAME, value='ibox')
    for i, el in enumerate(els):
        try:
            title = el.find_element(
                by=By.TAG_NAME,
                value='h5',
            ).text.lower()
        except:
            continue

        # Try to get cannabinoids data.
        if title == 'cannabinoids':
            totals = el.find_elements(
                by=By.TAG_NAME,
                value='compound-box',
            )
            for total in totals:
                value = total.find_element(
                    by=By.CLASS_NAME,
                    value='value',
                ).text
                units = total.find_element(
                    by=By.CLASS_NAME,
                    value='unit',
                ).text
                name = total.find_element(
                    by=By.CLASS_NAME,
                    value='name',
                ).text
                key = snake_case(name)
                obs[key] = convert_to_numeric(value)
                obs[f'{key}_units'] = units.replace('%', 'percent')

            # TODO: Generalize the below functionality into a function.

            # Get the cannabinoids totals.
            # Future work: Make the columns dynamic.
            columns = ['name', 'value', 'mg_g']
            table = el.find_element(by=By.TAG_NAME, value='table')
            rows = table.find_elements(by=By.TAG_NAME, value='tr')
            for row in rows[1:]:
                result = {'analysis': 'cannabinoids', 'units': 'percent'}
                cells = row.find_elements(by=By.TAG_NAME, value='td')
                for i, cell in enumerate(cells):
                    key = columns[i]
                    value = cell.get_attribute('textContent').strip()
                    if key == 'name':
                        result['key'] = parser.analytes.get(value, snake_case(value))
                    else:
                        value = value.replace('%', '').replace('mg/g', '').strip()
                    result[key] = convert_to_numeric(value)
                results.append(result)

        # Try to get terpene data.
        # Future work: Make the columns dynamic.
        if title == 'terpenes':
            columns = ['name', 'value', 'mg_g']
            try:
                table = el.find_element(by=By.TAG_NAME, value='table')
            except NoSuchElementException:
                continue
            rows = table.find_elements(by=By.TAG_NAME, value='tr')
            for row in rows[1:]:
                result = {'analysis': 'terpenes', 'units': 'percent'}
                cells = row.find_elements(by=By.TAG_NAME, value='td')
                for i, cell in enumerate(cells):
                    key = columns[i]
                    value = cell.get_attribute('textContent').strip()
                    if key == 'name':
                        result['key'] = parser.analytes.get(value, snake_case(value))
                    else:
                        value = value.replace('%', '').replace('mg/g', '').strip()
                    result[key] = convert_to_numeric(value)
                results.append(result)

            # Try to get predicted aromas.
            container = el.find_element(by=By.CLASS_NAME, value='row')
            aromas = container.text.split('\n')
            obs['predicted_aromas'] = [snake_case(x) for x in aromas]

        # Try to get screening data.
        if title == 'safety':
            obs['status'] = el.find_element(
                by=By.CLASS_NAME,
                value='sample-status'
            ).text
            table = el.find_element(by=By.TAG_NAME, value='table')
            rows = table.find_elements(by=By.TAG_NAME, value='tr')
            for row in rows[1:]:

                # Get the screening analysis.
                cells = row.find_elements(by=By.TAG_NAME, value='td')
                status = cells[1].get_attribute('textContent').strip()
                if status == 'Not Tested':
                    continue
                analysis = snake_case(cells[0].get_attribute('textContent'))
                obs[f'{analysis}_status'] = status.lower()
                analyses.append(analysis)

                # Scroll the row into view!
                parser.driver.execute_script('arguments[0].scrollIntoView(true);', row)

                # Click the row. and get all of the results from the modal!
                # Future work: Make the columns dynamic.
                columns = ['name', 'status', 'value', 'limit', 'loq']
                if row.get_attribute('class') == 'clickable-content':
                    row.click()
                    modal = parser.driver.find_element(
                        by=By.ID,
                        value='safety-modal-table'
                    )
                    modal_table = modal.find_element(
                        by=By.TAG_NAME,
                        value='tbody'
                    )
                    modal_rows = modal_table.find_elements(
                        by=By.TAG_NAME,
                        value='tr'
                    )
                    headers = modal.find_elements(
                        by=By.TAG_NAME,
                        value='th',
                    )
                    units = headers[-1].text.split('(')[-1].replace(')', '')
                    for modal_row in modal_rows:
                        result = {'analysis': analysis, 'units': units}
                        modal_cells = modal_row.find_elements(
                            by=By.TAG_NAME,
                            value='td'
                        )
                        for i, modal_cell in enumerate(modal_cells):
                            key = columns[i]
                            value = modal_cell.get_attribute(
                                'textContent'
                            ).strip()
                            if key == 'name':
                                result['key'] = parser.analytes.get(value, snake_case(value))
                            result[key] = convert_to_numeric(value)
                        results.append(result)

                    # Close the modal.
                    try:
                        button = parser.driver.find_element(
                            by=By.CLASS_NAME,
                            value='close',
                        )
                        button.click()
                    except ElementNotInteractableException:
                        continue
                    sleep(0.2) # Brief pause to give modal time to close.

        # Try to get lab data.
        producer = ''
        if title == 'order info':
            img = el.find_element(by=By.TAG_NAME, value='img')
            producer = el.find_element(
                by=By.CLASS_NAME,
                value='public-name',
            ).text
            license_el = el.find_element(
                by=By.TAG_NAME,
                value='confident-address',
            )
            lab = license_el.find_element(
                by=By.CLASS_NAME,
                value='address-name',
            ).text
            lab_phone = license_el.find_element(
                by=By.CLASS_NAME,
                value='address-phone',
            ).text.split(': ')[-1]
            lab_email = license_el.find_element(
                by=By.CLASS_NAME,
                value='address-email',
            ).text
            block = el.find_element(
                by=By.TAG_NAME,
                value='confident-address',
            ).text.split('\n')
            street = block[1]
            address = tuple(block[2].split(', '))
            obs['lab'] = lab
            obs['lab_address'] = f'{street} {", ".join(address)}'
            obs['lab_image_url'] = img.get_attribute('src')
            obs['lab_street'] = street
            obs['lab_city'] = address[0]
            obs['lab_state'], obs['lab_zipcode'] = tuple(address[-1].split(' '))
            obs['lab_phone'] = lab_phone
            obs['lab_email'] = lab_email
            obs['producer'] = producer

    # Return the sample with a freshly minted sample ID.
    obs = {**CONFIDENT_CANNABIS, **obs}
    obs['lab_results_url'] = url
    obs['analyses'] = json.dumps(list(set(analyses)))
    obs['results'] = json.dumps(results)
    obs['coa_algorithm_version'] = __version__
    obs['coa_parsed_at'] = datetime.now().isoformat()
    obs['results_hash'] = create_hash(results)
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=product_name,
        salt=producer,
    )
    obs['sample_hash'] = create_hash(obs)
    if not persist:
        parser.quit()
    return obs


def parse_cc_pdf(
        parser,
        doc: Any,
        **kwargs
    ) -> dict:
    """Parse a Confident Cannabis CoA PDF.
    Args:
        doc (str or PDF): A file path to a PDF or a pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """
    url = parser.find_pdf_qr_code_url(doc)
    return parse_cc_url(parser, url, **kwargs)


def parse_cc_coa(
        parser,
        doc: Any,
        **kwargs,
    ) -> dict:
    """Parse a Confident Cannabis CoA PDF or URL.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
    Returns:
        (dict): The sample data.
    """
    if isinstance(doc, str):
        if doc.startswith('http'):
            return parse_cc_url(parser, doc, **kwargs)
        elif doc.endswith('.pdf'):
            data = parse_cc_pdf(parser, doc, **kwargs)
        else:
            data = parse_cc_pdf(parser, doc, **kwargs)
    else:
        data = parse_cc_pdf(parser, doc, **kwargs)
    if isinstance(doc, str):
        data['coa_pdf'] = doc.replace('\\', '/').split('/')[-1]
    elif isinstance(doc, PDF):
        data['coa_pdf'] = doc.stream.name.replace('\\', '/').split('/')[-1]
    return data


if __name__ == '__main__':

    # Test Confident Cannabis CoAs parsing.
    from cannlytics.data.coas import CoADoc

    # Specify where your test data lives.
    DATA_DIR = '../../../tests/assets/coas'

    # # [✓] TEST: Parse a CoA URL.
    # cc_coa_url = 'https://share.confidentcannabis.com/samples/public/share/4ee67b54-be74-44e4-bb94-4f44d8294062'
    # parser = CoADoc()
    # data = parse_cc_url(parser, cc_coa_url)
    # assert data is not None

    # # [✓] TEST: Parse a CoA PDF.
    # cc_coa_pdf = f'{DATA_DIR}/Classic Jack.pdf'
    # parser = CoADoc()
    # data = parse_cc_pdf(parser, cc_coa_pdf)
    # assert data is not None

    # [✓] TEST: Parse a CoA PDF, figuring out that it's a CC CoA PDF.
    # directory = '../../../tests/assets/coas'
    # doc = f'{directory}/GDP.pdf'
    # parser = CoADoc()
    # lab = parser.identify_lims(doc)
    # assert lab == 'Confident Cannabis'
    # data = parse_cc_coa(parser, doc)
    # assert data is not None

    # [✓] TEST: Parse a CoA PDF, with safety screening but no terpenes.
    # directory = '../../../tests/assets/coas'
    # doc = f'{directory}/HGN_SOUR .pdf'
    # parser = CoADoc()
    # data = parse_cc_coa(parser, doc)
    # assert data is not None



"""
Cannlytics Web Data Initialization | Cannlytics
Copyright (c) 2023 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 7/2/2023
Updated: 7/2/2023
"""
try:
    from .web import (
        format_params,
        get_page_metadata,
        get_page_description,
        get_page_image,
        get_page_favicon,
        get_page_theme_color,
        get_page_phone_number,
        get_page_email,
        find_company_address,
        find_company_linkedin,
        find_company_url,
        download_google_drive_file,
    )

    __all__ = [
        format_params,
        get_page_metadata,
        get_page_description,
        get_page_image,
        get_page_favicon,
        get_page_theme_color,
        get_page_phone_number,
        get_page_email,
        find_company_address,
        find_company_linkedin,
        find_company_url,
        download_google_drive_file,
    ]
except ImportError:
    pass



"""
Analyte Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: API key data model.
"""
# Standard imports.
from dataclasses import dataclass
from datetime import datetime, timedelta

# Internal imports.
from .base import Model

@dataclass
class APIKey(Model):
    """A representation of an API key HMAC data."""
    _collection = 'admin/api/api_key_hmacs'
    expiration_at: datetime = datetime.now() + timedelta(365)
    name: str = ''
    # permissions: field(default_factory=list) = []
    uid: str = ''
    user_email: str = ''
    user_name: str = ''



"""
Module Setup | Cannlytics Engine
Copyright (c) 2021-2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Contact: <keegan@cannlytics.com>
Created: 1/21/2021
Updated: 6/1/2023
License: MIT <https://opensource.org/licenses/MIT>
"""
from setuptools import find_packages, setup

# Define the package version.
version = '0.0.15'

# Get the project description.
README = ''
with open('./cannlytics/readme.md', 'r', encoding='utf-8') as readme_file:
    README = readme_file.read()

# Specify requirements installed by `pip install cannlytics`.
REQUIREMENTS = []
with open('./cannlytics/requirements.txt', 'r') as f:
    REQUIREMENTS = [i[:-1] for i in f if i[0] != '#']

# Specify requirements for development.
dev_requirements = ['ocean-lib', 'xlwings']

# Specify requirements for setup.
setup_requirements = []

# Specify requirements for testing.
test_requirements = []

# Publish the module.
setup(
    author='Cannlytics',
    author_email='dev@cannlytics.com',
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.9',
    ],
    description='🔥 Cannlytics is a suite of tools that you can use to wrangle, standardize, and analyze cannabis data', #pylint: disable=line-too-long
    extras_require={
        'test': test_requirements,
        'dev': dev_requirements + test_requirements,
    },
    include_package_data=True,
    install_requires=REQUIREMENTS,
    keywords='cannlytics',
    license='MIT',
    long_description=README,
    long_description_content_type='text/markdown',
    name='cannlytics',
    packages=find_packages(include=['cannlytics', 'cannlytics.*']),
    python_requires='>=3.9',
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/cannlytics/cannlytics',
    version=version,
    zip_safe=False,
)



"""
CCRS Initialization | Cannlytics
Copyright (c) 2021-2023 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/6/2021
Updated: 1/3/2023
"""
from .ccrs import (
    anonymize,
    find_detections,
    format_lab_results,
    format_test_value,
    get_datafiles,
    merge_datasets,
    save_dataset,
    standardize_dataset,
    unzip_datafiles,
)
from .constants import (
    CCRS_ANALYSES,
    CCRS_ANALYTES,
    CCRS_DATASETS,
    CCRS_PLANT_GROWTH_STAGES,
    CCRS_PLANT_HARVEST_CYCLES,
    CCRS_PLANTS_SOURCES,
    CCRS_PLANT_STATES,
)

__all__ = [
    CCRS_ANALYSES,
    CCRS_ANALYTES,
    CCRS_DATASETS,
    CCRS_PLANT_GROWTH_STAGES,
    CCRS_PLANT_HARVEST_CYCLES,
    CCRS_PLANTS_SOURCES,
    CCRS_PLANT_STATES,
    anonymize,
    find_detections,
    format_lab_results,
    format_test_value,
    get_datafiles,
    merge_datasets,
    save_dataset,
    standardize_dataset,
    unzip_datafiles,
]



"""
Inventory Item Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Inventory item data model.
"""
# Standard imports.
from dataclasses import dataclass
from datetime import datetime

# Internal imports.
from .base import Model


@dataclass
class Item(Model):
    """Any physical inventory item that an organization may have at
    their facility(ies), such as cannabis flower, supplies, or
    instruments. Inventory can be assigned to a specific area to make
    locating the inventory item easier."""
    _collection = 'organizations/%s/inventory'
    admin_method: str = ''
    approved: str = ''
    approved_at: datetime = None
    area_id: str = ''
    area_name: str = ''
    category_name: str = ''
    category_type: str = ''
    description: str = ''
    dose: float = 0.0
    dose_number: int = 0
    dose_units: str = ''
    item_type: str = ''
    moved_at: datetime = None
    name: str = ''
    quantity: float = 0.0
    quantity_type: str = ''
    serving_size: float = 0.0
    supply_duration_days: int = 0
    status: str = ''
    strain_id: str = ''
    strain_name: str = ''
    units: str = ''
    volume: float = 0.0
    volume_units: str = ''
    weight: float = 0.0
    weight_units: str = ''



"""
CoADoc | Parse Confidence Analytics COAs
Copyright (c) 2022 Cannlytics

Authors:
    Candace O'Sullivan-Sutherland <https://github.com/candy-o>
    Keegan Skeate <https://github.com/keeganskeate>
Created: 11/26/2022
Updated: 12/2/2022
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description:

    Parse Confidence Analytics COA PDFs.

Data Points (✓):

    ✓ analyses
    ✓ methods
    - {analysis}_status
    ✓ date_harvested
    ✓ date_received
    ✓ date_tested
    ✓ external_id
    ✓ lab_id
    ✓ lab_results_url
    ✓ producer
    ✓ producer_street
    ✓ producer_city
    ✓ producer_state
    ✓ producer_zipcode
    ✓ producer_license_number
    ✓ producer_latitude
    ✓ producer_longitude
    ✓ product_id
    ✓ product_name
    ✓ product_type
    ✓ results
    ✓ results_hash
    ✓ sample_hash
    ✓ total_cannabinoids (calculated)
    ✓ total_cbd (Calculated as cbd + 0.877 * cbda)
    ✓ total_thc (Calculated as delta_9_thc + 0.877 * thca)
    ✓ total_terpenes (calculated)

"""
# Standard imports:
from datetime import datetime
import json
import os
from typing import Any, Optional, List

# External imports:
from datasets import load_dataset
import pandas as pd
import pdfplumber

# Internal imports:
from cannlytics import __version__
from cannlytics.data.data import create_hash, create_sample_id
from cannlytics.data.gis import search_for_address
from cannlytics.utils import (
    convert_to_numeric,
    download_file_from_url,
    snake_case,
)
from cannlytics.utils.constants import DECARB


# It is assumed that the lab has the following details.
CONFIDENCE = {
    'coa_algorithm': 'confidence.py',
    'coa_algorithm_entry_point': 'parse_confidence_coa',
    'lims': 'Confidence Analytics',
    'url': 'https://certs.conflabs.com',
    'lab': 'Confidence Analytics',
    'lab_website': 'https://conflabs.com',
    'lab_license_number': '7939083039',
    'lab_image_url': 'https://www.conflabs.com/wp-content/uploads/2022/06/CA_3_transparent_registered-2.png',
    'lab_address': '14797 NE 95th St, Redmond, WA 98502',
    'lab_street': '14797 NE 95th St',
    'lab_city': 'Redmond',
    'lab_county': 'King',
    'lab_state': 'WA',
    'lab_zipcode': '98052',
    'lab_latitude': 47.685890,
    'lab_longitude': -122.143640,
    'lab_phone': '(206) 743-8843',
    'lab_email': 'info@conflabs.com',
}

CONFIDENCE_COA = {
    'analyses': {
        'Cannabinoids': 'cannabinoids',
        'Foreign Matter': 'foreign_matter',
        'GC Pesticides': 'pesticides',
        'LC Pesticides': 'pesticides',
        'Microbes': 'microbes',
        'Mycotoxins': 'mycotoxins',
        'Terpenes': 'terpenes',
        'Water Activity': 'water_activity',
    },
    'fields': [
        'name',
        'analysis',
        'value',
        'limit',
        'units',
        'lod',
        'loq',
        'status',
        'date_tested',
    ],
}


def calculate_total_cannabinoids(
        results: List[dict],
        analysis: Optional[str] = 'cannabinoids',
        decarb: Optional[float] = DECARB,
        decimal_places: Optional[int] = 4,
    ) -> float:
    """Calculates total cannabinoids, applying a decarb rate to acidic
    cannabinoids, given a list of results.
    """
    total_cannabinoids = 0.0
    for result in results:
        if result['analysis'] == analysis:
            value = result['value']
            if isinstance(value, float):
                if result['key'].endswith('a'):
                    total_cannabinoids += decarb * value
                else:
                    total_cannabinoids += value
    return round(total_cannabinoids, decimal_places)


def calculate_total_terpenes(
        results: List[dict],
        analysis: Optional[str] = 'terpenes',
        decarb: Optional[float] = 1,
        decimal_places: Optional[int] = 4,
    ) -> float:
    """Calculates total terpenes given a list of results.
    """
    return calculate_total_cannabinoids(
        results,
        decarb=decarb,
        analysis=analysis,
        decimal_places=decimal_places,
    )


def calculate_total_cbd(
        results: List[dict],
        cbd_key: Optional[str] = 'cbd',
        cbda_key: Optional[str] = 'cbda',
        decarb: Optional[float] = DECARB,
    ) -> float:
    """Calculates total CBD given a list of results.
    """
    results_data = pd.DataFrame(results)
    try:
        cbd = results_data.loc[results_data['key'] == cbd_key].iloc[0]['value']
    except IndexError:
        cbd = 0
    try:
        cbda = results_data.loc[results_data['key'] == cbda_key].iloc[0]['value']
    except IndexError:
        cbda = 0
    cbd = convert_to_numeric(cbd)
    if not isinstance(cbd, float):
        cbd = 0
    cbda = convert_to_numeric(cbda)
    if isinstance(cbda, float):
        cbd += cbda * decarb
    return cbd


def calculate_total_thc(
        results: List[dict],
        thc_key: Optional[str] = 'delta_9_thc',
        thca_key: Optional[str] = 'thca',
        decarb: Optional[float] = DECARB,
    ) -> float:
    """Calculates total THC given a list of results.
    """
    results_data = pd.DataFrame(results)
    try:
        thc = results_data.loc[results_data['key'] == thc_key].iloc[0]['value']
    except IndexError:
        thc = 0
    try:
        thca = results_data.loc[results_data['key'] == thca_key].iloc[0]['value']
    except IndexError:
        thca = 0
    thc = convert_to_numeric(thc)
    if not isinstance(thc, float):
        thc = 0
    thca = convert_to_numeric(thca)
    if isinstance(thca, float):
        thc += thca * decarb
    return thc


def parse_confidence_pdf(
        parser,
        doc: Any,
        coa_pdf: Optional[str] = '',
        **kwargs,
    ) -> dict:
    """Parse a Confidence Analytics COA PDF.
    Args:
        parser (CoADoc): A CoADoc parsing client.
        doc (str or PDF): A PDF file path or pdfplumber PDF.
        coa_pdf (str): A filename to use for the `coa_pdf` field (optional).
    Returns:
        (dict): The sample data.
    """
    # Read the PDF.
    obs = {}
    if isinstance(doc, str):
        report = pdfplumber.open(doc)
        obs['coa_pdf'] = coa_pdf or doc.replace('\\', '/').split('/')[-1]
    else:
        report = doc
        obs['coa_pdf'] = coa_pdf or report.stream.name.replace('\\', '/').split('/')[-1]

    # Get the lab data.
    front_page = report.pages[0]
    front_page_text = front_page.extract_text()
    front_page_lines = front_page_text.split('\n')
    for i, line in enumerate(front_page_lines):

        # Get the sample number as the lab ID.
        if 'Sample #' in line:
            obs['lab_id'] = line.split('Sample # ')[-1]

        # Get the product name, product ID, and date received.
        elif 'Sample Name:' in line:
            next_line = front_page_lines[i + 1]
            parts = next_line.split(' ')
            obs['date_received'] = parts[-1]
            obs['product_id'] = parts[-2]
            obs['product_name'] = ' '.join(parts[0:-2])

        # Get the product type, external ID, and date tested.
        elif 'Type:' in line:
            next_line = front_page_lines[i + 1]
            parts = next_line.split(' ')
            obs['date_tested'] = parts[-1]
            obs['external_id'] = parts[-2]
            obs['product_type'] = ' '.join(parts[0:-2])

        # Get the producer.
        elif 'Origination:' in line:
            next_line = front_page_lines[i + 1]
            parts = next_line.split(',')[0].split(' ')
            obs['producer_license_number'] = parts[-3]
            obs['producer'] = ' '.join(parts[0:-3])
        
        # Get the producer address.
        elif 'Address:' in line:
            next_line = front_page_lines[i + 1]
            parts = next_line.split(' Date of Harvest:')[0].split(' ')
            obs['producer_street'] = ' '.join(parts[0:-1])

            # Get the city, zip code, and date harvested from the next line.
            address_line = front_page_lines[i + 2]
            address_parts = address_line.split(',')
            zip_parts = address_parts[-1].strip().split(' ')
            obs['producer_city'] = address_parts[0].title()
            obs['producer_state'] = zip_parts[0]
            obs['producer_zipcode'] = zip_parts[1]
            obs['producer_address'] = ', '.join([
                obs['producer_street'],
                obs['producer_city'],
                obs['producer_state'] + ' ' + obs['producer_zipcode'],
            ])
            if '(not provided)' in address_line:
                obs['date_harvested'] = None
            else:
                obs['date_harvested'] = zip_parts[-1]
        
        # Get the sum of cannabinoids.
        elif 'Total Canna. (raw sum):' in line:
            value = float(line.split('%')[0].split(':')[-1])
            obs['sum_cannabinoids'] = round(value, 3)

        # TODO: Get the analysis statuses.
        # 'Foreign Matter + Seeds:PASS Microbes:PASS Pesticides:PASS',
        # 'Water Activity:PASS Mycotoxins:PASS Heavy Metals:NE',
        # 'Residual Solvents:NE   '

    # Get all of the lines of the document.
    all_lines = []
    for page in report.pages[1:]:
        all_lines.extend(page.extract_text().split('\n'))

    # Aggregate all of the analyte tables.
    analyte_lines = []
    header = False
    for line in all_lines:
        if 'Analyte Name' in line:
            header = True
        elif 'Document Created' in line:
            header = False
        elif '[ END OF ANALYTE TABLE ]' in line:
            break
        elif header:
            analyte_lines.append(line)

    # Remove blank lines from the analyte tables.
    analyte_lines = [x for x in analyte_lines if x.strip()]

    # Format the results.
    analyses, results = [], []
    standard_analyses = CONFIDENCE_COA['analyses']
    standard_fields = CONFIDENCE_COA['fields']
    analysis_names = list(standard_analyses.keys())
    analysis_keys = list(standard_analyses.values())
    for line in analyte_lines:

        # Identify the analysis.
        values = line.replace('< MRL', '<LOQ')
        for n, analysis in enumerate(analysis_names):
            if analysis in values:
                analysis_key = analysis_keys[n]
                values = values.replace(analysis, analysis_key)
                analyses.append(analysis_key)
                break

        # Remove extra spaces.
        values = values.split(' ')
        values = [x for x in values if x.strip()]

        # Parse the result for the analyte.
        result = {}
        for k, field in enumerate(reversed(standard_fields)):
            if field == 'name':
                name = ' '.join(values[0:-8])
                name = name.rstrip('3')
                key = snake_case(name)
                analyte = parser.analytes.get(key, key)
                result['name'] = name
                result['key'] = analyte
            else:
                result[field] = convert_to_numeric(values[-k - 1])

        # Skip totals.
        if analyte == 'raw' or analyte == 'total' or not analyte:
            continue
        
        # Record the analyte result.
        results.append(result)

    # Convert terpenes to percentages.
    for i, result in enumerate(results):
        if result['analysis'] == 'terpenes':
            result['units'] = 'percent'
            try:
                result['value'] = result['value'] / 10_000
            except TypeError:
                pass
            results[i] = result

    # Calculate total cannabinoids, terpenes, CBD, and THC.
    obs['total_cannabinoids'] = calculate_total_cannabinoids(results)
    obs['total_terpenes'] = calculate_total_terpenes(results)
    obs['total_cbd'] = calculate_total_cbd(results)
    obs['total_thc'] = calculate_total_thc(results)

    # Get the producer's latitude and longitude from the
    # `cannabis_licenses` dataset or by geocoding their address.
    wa_licenses = load_dataset('cannlytics/cannabis_licenses', 'wa')
    licenses = wa_licenses['data'].to_pandas()
    criterion = licenses['license_number'].str.contains(obs['producer_license_number'])
    match = licenses.loc[criterion]
    if len(match):
        licensee = match.iloc[0]
        obs['producer_county'] = licensee['premise_county']
        obs['producer_latitude'] = licensee['premise_latitude']
        obs['producer_longitude'] = licensee['premise_longitude']
    else:
        try:
            google_maps_api_key = os.environ['GOOGLE_MAPS_API_KEY']
            location = search_for_address(
                obs['producer_address'],
                api_key=google_maps_api_key
            )
            for key, value in location.items():
                obs[f'producer_{key}'] = value
        except:
            print("""Set `GOOGLE_MAPS_API_KEY` environment variable to
            get the producer's latitude and longitude.""")
            pass

    # Get the methods.
    methods = []
    lines = report.pages[-1].extract_text().split('\n')
    header = False
    for line in lines:
        if 'Analytical Methods Used' in line:
            header = True
        elif header:
            method = line.strip()
            if method:
                methods.append(method)
            else:
                break

    # Get the lab results URL from the QR code.
    obs['lab_results_url'] = parser.find_pdf_qr_code_url(front_page)
    
    # Close the report.
    report.close()

    # Turn dates to ISO format.
    date_columns = [x for x in obs.keys() if x.startswith('date')]
    for date_column in date_columns:
        try:
            obs[date_column] = pd.to_datetime(obs[date_column]).isoformat()
        except:
            pass

    # Finish data collection with a freshly minted sample ID.
    obs = {**CONFIDENCE, **obs}
    obs['analyses'] = json.dumps(list(set(analyses)))
    obs['coa_algorithm_version'] = __version__
    obs['coa_parsed_at'] = datetime.now().isoformat()
    obs['methods'] = json.dumps(methods)
    obs['results'] = results
    obs['results_hash'] = create_hash(results)
    obs['sample_id'] = create_sample_id(
        private_key=json.dumps(results),
        public_key=obs['product_name'],
        salt=obs.get('producer', obs.get('date_tested', 'cannlytics.eth')),
    )
    obs['sample_hash'] = create_hash(obs)
    return obs


def parse_confidence_url(
        parser,
        url: str,
        temp_path: Optional[str] = '/tmp',
        **kwargs
    ) -> dict:
    """Parse a Confidence Analytics COA URL.
    Args:
        url (str): The COA URL.
        headers (Any): Optional headers for standardization.
        max_delay (float): The maximum number of seconds to wait
            for the page to load.
        persist (bool): Whether to persist the driver.
            The default is `False`. If you do persist
            the driver, then make sure to call `quit`
            when you are finished.
    Returns:
        (dict): The sample data.
    """
    # Download the PDF.
    coa_pdf = download_file_from_url(url, temp_path, ext='.pdf')

    # Extract the data with the PDF parsing algorithm.
    data = parse_confidence_pdf(parser, coa_pdf)

    # Record the URL.
    data['lab_results_url'] = url
    return data


def parse_confidence_coa(
        parser,
        doc: Any,
        **kwargs,
    ) -> dict:
    """Parse a Confidence Analytics COA PDF or URL.
    Args:
        doc (str or PDF): A PDF file path or pdfplumber PDF.
        kwargs (arguments): Arguments to pass to the parsing algorithms.
    Returns:
        (dict): The sample data.
    """
    if isinstance(doc, str):
        if doc.startswith('https'):
            return parse_confidence_url(parser, doc, **kwargs)
    return parse_confidence_pdf(parser, doc, **kwargs)


# === Tests ===
if __name__ == '__main__':

    from cannlytics.data.coas import CoADoc
    from dotenv import dotenv_values

    # Set a Google Maps API key.
    config = dotenv_values('../../../.env')
    os.environ['GOOGLE_MAPS_API_KEY'] = config['GOOGLE_MAPS_API_KEY']

    # Specify testing constants.
    coa_url = 'https://certs.conflabs.com/full/WA-kXIRbGqnLVBI-WA-221015-052.pdf'
    doc = '../../../tests/assets/coas/confidence-analytics/13232582566688421-Flower-Dirty-Banana-Breath-WA-221015-052.pdf'
    temp_path = '../../../tests/assets/coas/tmp'

    # [✓] TEST: Identify LIMS.
    parser = CoADoc()
    lims = parser.identify_lims(doc, lims={'Confidence Analytics': CONFIDENCE})
    assert lims == 'Confidence Analytics'

    # [✓] TEST: Parse COA PDF.
    parser = CoADoc()
    data = parse_confidence_pdf(parser, doc)
    assert data is not None

    # [✓] TEST: Parse COA URL.
    parser = CoADoc()
    data = parse_confidence_url(parser, coa_url, temp_path=temp_path)
    assert data is not None

    # [✓] TEST: Parse COA URL and PDF ambiguously.
    parser = CoADoc()
    coas = [coa_url, doc]
    for coa in coas:
        data = parser.parse(coa, temp_path=temp_path)
        assert data is not None
    print('✓ Completed Confidence Analytics COA parsing tests.')



"""
Template Model | Cannlytics
Copyright (c) 2021-2022 Cannlytics and Cannlytics Contributors

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 11/5/2021
Updated: 11/5/2021
License: <https://github.com/cannlytics/cannlytics/blob/main/LICENSE>

Description: Template data model.
"""
# Standard imports.
from dataclasses import dataclass
from datetime import datetime

# Internal imports.
from .base import Model


@dataclass
class Template(Model):
    """Templates for generating documents, such as invoices and
    certificates."""
    _collection = 'organizations/%s/transfers'
    status: str = ''
    storage_ref: str = ''
    version: str = ''



"""
Cannlytics Sales Data Initialization | Cannlytics
Copyright (c) 2022 Cannlytics

Authors: Keegan Skeate <https://github.com/keeganskeate>
Created: 5/13/2023
Updated: 5/13/2023
"""
try:
    from .receipts_ai import ReceiptsParser
    __all__ = [
        ReceiptsParser,
    ]
except ImportError:
    pass
